#include <v_lib.h>

/*
 * Binary tree image decomposition
 */

typedef struct
{
	u32_t left, top, right, bottom;
	vec_t value;
} btdecomp_node_t, *btdecomp_node_p;

void
matMinMax(matrix_p data,
	u32_t left,u32_t top,u32_t right,u32_t bottom,
	vec_t *min_ret,
	vec_t *max_ret)
{
	u32_t r, c;
	vec_t min, max, *row;
	min = max = data->cell[top][left];
	for(r = top;r <= bottom;r++)
	{
		row = data->cell[r];
		for(c = left;c <= right;c++)
		{
			if(min > row[c])
				min = row[c];
			else if(max < row[c])
				max = row[c];
		}
	}
	if(min_ret)
		*min_ret = min;
	if(max_ret)
		*max_ret = max;
}

u32_t matFindBestVSplit(matrix_p data,
	u32_t left,u32_t top,u32_t right,u32_t bottom,
	vec_t *max_diff_ret)
{
	u32_t split, best_split;
	vec_t umin, umax, lmin, lmax, diff, max_diff = 0;
	for(best_split = split = top+1;split < bottom;split++)
	{
		matMinMax(data,left,top,right,split,&umin,&umax);
		matMinMax(data,left,split,right,bottom,&lmin,&lmax);
		/* absolute difference between averages of upper and lower submatrices */
		diff = fabs(umin + umax - lmin - lmax) * (vec_t)0.5;
		if(max_diff < diff)
		{
			max_diff = diff;
			best_split = split;
		}
	}
	if(max_diff_ret)
		*max_diff_ret = max_diff;
	return best_split;
}

u32_t matFindBestHSplit(matrix_p data,
	u32_t left,u32_t top,u32_t right,u32_t bottom,
	vec_t *max_diff_ret)
{
	u32_t split, best_split;
	vec_t lmin, lmax, rmin, rmax, diff, max_diff = 0;
	for(best_split = split = left+1;split < right;split++)
	{
		matMinMax(data,left,top,split,bottom,&lmin,&lmax);
		matMinMax(data,split,top,right,bottom,&rmin,&rmax);
		/* absolute difference between averages of left and right submatrices */
		diff = fabs(lmin + lmax - rmin - rmax) * (vec_t)0.5;
		if(max_diff < diff)
		{
			max_diff = diff;
			best_split = split;
		}
	}
	if(max_diff_ret)
		*max_diff_ret = max_diff;
	return best_split;
}

btree_p V_CALL
bt_img_decomp_r(matrix_p data,
	u32_t left,u32_t top,u32_t right,u32_t bottom,
	vec_t thresold)
{
	u32_t vsplit, hsplit;
	vec_t min, max;
	btree_p node;
	btdecomp_node_t node_data;
	vboolean vspan = left == right; /* vertical span of pixels */
	vboolean hspan = top == bottom; /* horizontal span */
	node_data.left = left;
	node_data.top = top;
	node_data.right = right;
	node_data.bottom = bottom;
	if(hspan && vspan)
	{
		/* one pixel: leaf */
		node_data.value = data->cell[top][left];
		return btStore(&node_data,sizeof(node_data),NULL,0);
	}
	matMinMax(data,left,top,right,bottom,&min,&max);
	node_data.value = (min + max) * (vec_t)0.5;
	node = btStore(&node_data,sizeof(node_data),NULL,0);
	if((max - min) < thresold) /* difference too small: leaf */
		return node;
#if 0
	if(vspan)
	{
		vsplit = matFindBestVSplit(data,left,top,right,bottom,NULL);
		node->children[0] = bt_img_decomp_r(data,left,top,right,vsplit,thresold);
		node->children[0]->parent = node;
		node->children[1] = bt_img_decomp_r(data,left,vsplit,right,bottom,thresold);
		node->children[1]->parent = node;
		return node;
	}
	if(hspan)
	{
		hsplit = matFindBestHSplit(data,left,top,right,bottom,NULL);
		node->children[0] = bt_img_decomp_r(data,left,top,hsplit,bottom);
		node->children[0]->parent = node;
		node->children[1] = bt_img_decomp_r(data,hsplit+1,top,right,bottom);
		node->children[1]->parent = node;
		return node;
	}
#endif
	/* choose split direction */
	if((right-left) > (bottom-top))
	{
		/* split into left and right submatrices */
		hsplit = left + ((right - left) / 2);
		node->children[0] = bt_img_decomp_r(data,left,top,hsplit,bottom,thresold);
		node->children[0]->parent = node;
		node->children[1] = bt_img_decomp_r(data,hsplit+1,top,right,bottom,thresold);
		node->children[1]->parent = node;
	}
	else
	{
		/* split into upper and lower submatrices */
		vsplit = top + ((bottom - top) / 2);
		node->children[0] = bt_img_decomp_r(data,left,top,right,vsplit,thresold);
		node->children[0]->parent = node;
		node->children[1] = bt_img_decomp_r(data,left,vsplit+1,right,bottom,thresold);
		node->children[1]->parent = node;
	}
	return node;
}


void
matSetRect(matrix_p data,
	u32_t left,u32_t top,u32_t right,u32_t bottom,
	vec_t value)
{
	u32_t r, c;
	vec_t *row;
	for(r = top;r <= bottom;r++)
	{
		row = data->cell[r];
		for(c = left;c <= right;c++)
			row[c] = value;
	}
}

void
bt_img_draw_decomp_r(matrix_p surface,btree_p root)
{
	btdecomp_node_p nd;
	if(!root->children[0]) /* leaf */
	{
		nd = (btdecomp_node_p)root->data;
		matSetRect(surface,nd->left,nd->top,nd->right,nd->bottom,nd->value);
		return;
	}
	bt_img_draw_decomp_r(surface,root->children[0]);
	bt_img_draw_decomp_r(surface,root->children[1]);
}

void
bt_img_draw_gray_decomp_r(matrix_p surface,btree_p root)
{
	vec_t area, full_area = surface->cols * surface->rows;
	btdecomp_node_p nd;
	if(!root->children[0]) /* leaf */
	{
		nd = (btdecomp_node_p)root->data;
		area = (nd->bottom - nd->top + 1) * (nd->right - nd->left + 1);
		area = (vec_t)(rand() & 255) / 255;
		matSetRect(surface,nd->left,nd->top,nd->right,nd->bottom,area);
		return;
	}
	bt_img_draw_gray_decomp_r(surface,root->children[0]);
	bt_img_draw_gray_decomp_r(surface,root->children[1]);
}

typedef struct btdecomp_leaf_info_s
{
	u32_t depth;
	u32_t left,top,right,bottom;
	u32_t area;
	vec_t value;
}	btdecomp_leaf_info_t, *btdecomp_leaf_info_p;

void
bt_img_describe_r(bstack_p info_stack,btree_p root,u32_t depth)
{
	btdecomp_node_p nd;
	btdecomp_leaf_info_t info;
	if(!root->children[0]) /* leaf */
	{
		nd = (btdecomp_node_p)root->data;
		info.depth = depth;
		info.left = nd->left;
		info.top = nd->top;
		info.right = nd->right;
		info.bottom = nd->bottom;
		info.area = (info.right - info.left + 1) * (info.bottom - info.top + 1);
		info.value = nd->value;
		stackPush(info_stack,&info);
		return;
	}
	depth++;
	bt_img_describe_r(info_stack,root->children[0],depth);
	bt_img_describe_r(info_stack,root->children[1],depth);
}

bstack_p 
bt_img_describe(btree_p root)
{
	bstack_p info_stack = (bstack_p)memAlloc(sizeof(bstack_t));
	stackInit(info_stack,sizeof(btdecomp_leaf_info_t));
	bt_img_describe_r(info_stack,root,0);
	return info_stack;
}

/*
 * ===============================================================
 */

matrix_p load_image(char *name)
{
	u32_t w,h,i,j;
	matrix_p m;
	vec_t scale = 1.0 / 3 / 256;
	u8_t *ptr,*img = jpgLoad(name,&w,&h);
	if(!img)
		return NULL;
	m = matAlloc(w,h);
	ptr = img;
	for(i=0;i<m->rows;i++)
		for(j=0;j<m->cols;j++)
		{
			m->cell[i][j] = (ptr[0]+ptr[1]+ptr[2]) * scale;
			ptr += 4;
		}
	memFree(img);
	return m;
}

void save_image(char *name,matrix_p src)
{
	u32_t i,j;
	vec_t val;
	u8_t *ptr, *img = (u8_p)memAllocZ(src->rows * src->cols * 4);
	if(!img)
		return;
	ptr = img;
	for(i=0;i < src->rows;i++)
		for(j=0;j < src->cols;j++)
		{
			val = src->cell[i][j] * 255;
			if(val < 0)
				val = 0;
			else if(val > 255)
				val = 255;
			ptr[0] = ptr[1] = ptr[2] = (u8_t)val;
			ptr += 4;
		}
	bmpSave(name,src->cols,src->rows,img);
	memFree(img);
}

int main()
{
	btree_p root;
	matrix_p img;
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	
	img = load_image("container.jpg");
	assert(img);

	/* save_image("out0.bmp",img); */

	root = bt_img_decomp_r(img,0,0,img->cols-1,img->rows-1,0.5);
	bt_img_draw_gray_decomp_r(img,root);
	
	/*
	{
		bstack_p info = bt_img_describe(root);
		stackClear(info);
		memFree(info);
	}
	*/
	btDeleteAllNodes(root,NULL);

	save_image("out1.bmp",img);
    memFree(img);

	memPrintStats();
	return 0;
}
