#include <v_lib.h>

/*
 * This is a small subset of USL ver. 0.01.
 * USL stands for Ugly Scripting Language.
 */

#define STR_COMP        strcasecmp
#define STR_EQ(s1,s2)   (!STR_COMP(s1,s2))
#define MAX_ID          64

#ifdef  CHECKED_BUILD

#define PARSE_ERROR(scanner,desc) \
	conlPrintf(LOG_ERROR, \
		"%s(): error in \"%s\", line %u: %s (got \"%s\").\n", \
		caller_name,lexSourceName(scanner), \
		lexLastLine(scanner),desc,strSafe(lexLast(scanner)))

#define SYN_EXPECTED(scanner,desc) \
	conlPrintf(LOG_ERROR, \
		"%s(): error in \"%s\", line %u: expected \"%s\", got \"%s\".\n", \
		caller_name,lexSourceName(scanner), \
		lexLastLine(scanner),desc,strSafe(lexLast(scanner)))

#define SYN_UNDECLARED(scanner,desc) \
	conlPrintf(LOG_ERROR, \
		"%s(): error in \"%s\", line %u: \"%s\" undeclared.\n", \
		caller_name,lexSourceName(scanner),lexLastLine(scanner),desc)

#else   /* !CHECKED_BUILD */

#define PARSE_ERROR(scanner,desc)
#define SYN_EXPECTED(scanner,desc)
#define SYN_UNDECLARED(scanner,id_name)

#endif  /* !CHECKED_BUILD */

#define SYN_EXPECTED_EXPRESSION(scanner)    SYN_EXPECTED(scanner,"expression")

/*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Part 1: expressons.
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

/*
 * The first position in variables' array reserved for exit code.
 */
#define EXIT_CODE_OFFSET    0
#define EXIT_CODE_NAME      "%" /* mimic Maple syntax */

typedef struct et_built_in_var_s
{
	char    *var_name;
	int     var_offset;
	char    *var_desc;
}   et_built_in_var_t, *et_built_in_var_p;

typedef struct et_built_in_const_s
{
	char    *const_name;
	vec_t   const_value;
	char    *const_desc;
}   et_built_in_const_t, *et_built_in_const_p;

V_LOCAL et_built_in_var_t known_built_in_vars[] =
{
	{EXIT_CODE_NAME,EXIT_CODE_OFFSET,"exit code of last operator"},
	{"$CELL",EXIT_CODE_OFFSET+1,"value of current cell (valid in 'mat_trans')"},
	{"$ROW",EXIT_CODE_OFFSET+2,"current row"},
	{"$COLUMN",EXIT_CODE_OFFSET+3,"current column"},
	{"$MAX_ROW",EXIT_CODE_OFFSET+4,"maximal index of row (num_rows - 1)"},
	{"$MAX_COLUMN",EXIT_CODE_OFFSET+5,"maximal index of column (num_cols - 1)"},
	{"$RE",EXIT_CODE_OFFSET+6,"real part of current FFT cell (valid for 'cart_trans')"},
	{"$IM",EXIT_CODE_OFFSET+7,"imaginary part of current FFT cell (valid for 'cart_trans')"},
	{"$MAGNITUDE",EXIT_CODE_OFFSET+8,"magnitude of current FFT cell (valid for 'polar_trans')"},
	{"$PHASE",EXIT_CODE_OFFSET+9,"phase of current FFT cell (valid for 'polar_trans')"},
	{NULL,-1,NULL}
};

V_LOCAL et_built_in_const_t known_built_in_consts[] =
{
	{"Pi",M_PI,"built-in: Pi"},
	{NULL,0,NULL}
};

int
etFirstBuiltInVariableOffset(void)
{
	et_built_in_var_p ptr = known_built_in_vars;
	if(ptr->var_name)
		return ptr->var_offset;
	return -1;
}

int
etLastBuiltInVariableOffset(void)
{
	et_built_in_var_p ptr = known_built_in_vars;
	if(!ptr->var_name)
		return -1;
	while(ptr->var_name)
		ptr++;
	ptr--;
	return ptr->var_offset;
}

vboolean
etFindBuiltInVariable(char *var_name,int *idx_ret)
{
	et_built_in_var_p ptr = known_built_in_vars;
	while(ptr->var_name)
	{
		if(STR_EQ(var_name,ptr->var_name))
		{
			if(idx_ret)
				*idx_ret = ptr->var_offset;
			return true;
		}
		ptr++;
	}
	return false;
}

vboolean
etFindBuiltInConstant(char *const_name,vec_t *const_ret)
{
	et_built_in_const_p ptr = known_built_in_consts;
	while(ptr->const_name)
	{
		if(STR_EQ(const_name,ptr->const_name))
		{
			if(const_ret)
				*const_ret = ptr->const_value;
			return true;
		}
		ptr++;
	}
	return false;
}

/* -------------------------------------------------------------------- */

V_LOCAL void
etNewIdentifier(u8_stack_p var_ids,char *var_name)
{

	char id[MAX_ID];
	memZero(id,MAX_ID);
	strCopy(id,var_name,MAX_ID);
	u8sPushBuffer(var_ids,(u8_p)id,MAX_ID);
	return;
}

V_LOCAL vboolean
etInitIdentifiers(u8_stack_p var_ids)
{
	int idx = 0;
	et_built_in_var_p ptr = known_built_in_vars;
	u8sInit(var_ids);
	/* add built-in variables */
	while(ptr->var_name)
	{
		/* _must_ be ordered */
		if(idx != ptr->var_offset)
		{
#ifdef  CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"etInitIdentifiers(): unordered built-in variable '%s'.\n",
				ptr->var_name);
#endif  /* CHECKED_BUILD */
			u8sClear(var_ids);
			return false;
		}
		etNewIdentifier(var_ids,ptr->var_name);
		ptr++;
		idx++;
	}
	return true;
}

V_LOCAL vboolean
etFindIdentifier(char *var_name,u8_stack_p var_ids,int *idx_ret)
{
	char *ptr;
	u32_t i, len;
	len = u8sCount(var_ids) / MAX_ID;
	ptr = (char*)u8sGetStorage(var_ids);
	for(i = 0;i < len;i++,ptr += MAX_ID)
	{
		if(STR_EQ(var_name,ptr))
		{
			if(idx_ret)
				*idx_ret = i;
			return true;
		}
	}
	return false;
}

/*
 * Node of expression tree
 */
struct et_node_s;

typedef vec_t (*et_op_t)(struct et_node_s *self,vec_t *vars);

typedef struct et_node_s
{
	union
	{
		int     idx;
		vec_t   constant;
	} var;
	et_op_t     op_ptr;
	/*
	 * The third branch used only in ternary '?' statement
	 */
	struct      et_node_s *left, *right, *control;
}   et_node_t, *et_node_p;

#define ET_EVAL(root,vars)  ((root)->op_ptr(root,vars))

/*
 * Constants and variables
 */
#include <../source/PAL/argsused.h>
V_LOCAL vec_t
op_const(et_node_p root,vec_t *vars)
{
	return root->var.constant;
}

V_LOCAL vec_t
op_var(et_node_p root,vec_t *vars)
{
	return vars[root->var.idx];
}

/*
 * Unary operations
 */
V_LOCAL vec_t
op_neg(et_node_p root,vec_t *vars)
{
	return -ET_EVAL(root->left,vars);
}

V_LOCAL vec_t
op_abs(et_node_p root,vec_t *vars)
{
	return v_abs(ET_EVAL(root->left,vars));
}

V_LOCAL vec_t
op_sin(et_node_p root,vec_t *vars)
{
	return (vec_t)sin(ET_EVAL(root->left,vars));
}

V_LOCAL vec_t
op_cos(et_node_p root,vec_t *vars)
{
	return (vec_t)cos(ET_EVAL(root->left,vars));
}

V_LOCAL vec_t
op_exp(et_node_p root,vec_t *vars)
{
	return (vec_t)exp(ET_EVAL(root->left,vars));
}

V_LOCAL vec_t
op_ln(et_node_p root,vec_t *vars)
{
	return (vec_t)log(ET_EVAL(root->left,vars));
}

/*
 * Binary operations
 */
V_LOCAL vec_t
op_add(et_node_p root,vec_t *vars)
{
	return ET_EVAL(root->left,vars) + ET_EVAL(root->right,vars);
}

V_LOCAL vec_t
op_sub(et_node_p root,vec_t *vars)
{
	return ET_EVAL(root->left,vars) - ET_EVAL(root->right,vars);
}

V_LOCAL vec_t
op_mul(et_node_p root,vec_t *vars)
{
	return ET_EVAL(root->left,vars) * ET_EVAL(root->right,vars);
}

V_LOCAL vec_t
op_div(et_node_p root,vec_t *vars)
{
	return ET_EVAL(root->left,vars) / ET_EVAL(root->right,vars);
}

V_LOCAL vec_t
op_mod(et_node_p root,vec_t *vars)
{
	return (vec_t)fmod(ET_EVAL(root->left,vars),ET_EVAL(root->right,vars));
}

V_LOCAL vec_t
op_pow(et_node_p root,vec_t *vars)
{
	return pow(ET_EVAL(root->left,vars),ET_EVAL(root->right,vars));
}

/*
 * Pseudo-boolean operations.
 * Returns:
 *  - 'false': _less_ than 0.5 (currently 0);
 *  - 'true': _greater_ than 0.5 (currently 1).
 */
V_LOCAL vec_t
op_less(et_node_p root,vec_t *vars)
{
	return (ET_EVAL(root->left,vars) < ET_EVAL(root->right,vars)) ?
		1 : 0;
}

V_LOCAL vec_t
op_greater(et_node_p root,vec_t *vars)
{
	return (ET_EVAL(root->left,vars) > ET_EVAL(root->right,vars)) ?
		1 : 0;
}

V_LOCAL vec_t
op_equal(et_node_p root,vec_t *vars)
{
	return IsZero(ET_EVAL(root->left,vars) - ET_EVAL(root->right,vars)) ?
		1 : 0;
}

V_LOCAL vec_t
op_not_equal(et_node_p root,vec_t *vars)
{
	return IsZero(ET_EVAL(root->left,vars) - ET_EVAL(root->right,vars)) ?
		0 : 1;
}

/*
 * Ternary operation
 *  control <= 0.5: select 'left'; else select 'right'.
 */
V_LOCAL vec_t
op_select(et_node_p root,vec_t *vars)
{
	return (ET_EVAL(root->control,vars) <= (vec_t)0.5) ? 
		ET_EVAL(root->left,vars) : ET_EVAL(root->right,vars);
}

/*
 * List of known operations
 */

typedef struct et_op_entry_s
{
	char        *op_name;
	et_op_t     op_ptr;
	char        *op_desc;
}   et_op_entry_t, *et_op_entry_p;

V_LOCAL et_op_entry_t known_unary_operations[] =
{
	{"neg",op_neg,"negates tos[0] (unary minus)"},
	{"abs",op_abs,"absolute value of tos[0]"},
	{"sin",op_sin,"sine of tos[0]"},
	{"cos",op_cos,"cosine of tos[0]"},
	{"exp",op_exp,"exponent of tos[0]"},
	{"ln",op_ln,"natural logarithm of tos[0]"},
	{NULL,NULL,NULL}
};

V_LOCAL et_op_entry_t known_binary_operations[] =
{
	{"+",op_add,"adds tos[1] to tos[0]"},
	{"-",op_sub,"subtracts tos[0] form tos[1]"},
	{"*",op_mul,"multiplies tos[1] by tos[0]"},
	{"/",op_div,"divides tos[1] by tos[0]"},
	{"%",op_mod,"remainder of tos[1] / tos[0]"},
	{"^",op_pow,"raises tos[1] to the power of tos[0]"},
	{"<",op_less,"returns 1 if tos[1] < tos[0]"},
	{">",op_greater,"returns 1 if tos[1] > tos[0]"},
	{"==",op_equal,"returns 1 if |tos[1] - tos[0]| < FLOAT_ZERO_EPSILON"},
	{"!=",op_not_equal,"returns 1 if |tos[1] - tos[0]| > FLOAT_ZERO_EPSILON"},
	{NULL,NULL,NULL}
};

V_LOCAL et_op_entry_t known_ternary_operations[] =
{
	{"?",op_select,"if(tos[0] <= 0.5) return tos[1]; else return tos[2]"},
	{NULL,NULL,NULL}
};

et_op_t
etFindUnaryOp(char *op_name)
{
	et_op_entry_p ptr = known_unary_operations;
	while(ptr->op_name)
	{
		if(STR_EQ(op_name,ptr->op_name))
			break;
		ptr++;
	}
	return ptr->op_ptr;
}

et_op_t
etFindBinaryOp(char *op_name)
{
	et_op_entry_p ptr = known_binary_operations;
	while(ptr->op_name)
	{
		if(STR_EQ(op_name,ptr->op_name))
			break;
		ptr++;
	}
	return ptr->op_ptr;
}

et_op_t
etFindTernaryOp(char *op_name)
{
	et_op_entry_p ptr = known_ternary_operations;
	while(ptr->op_name)
	{
		if(STR_EQ(op_name,ptr->op_name))
			break;
		ptr++;
	}
	return ptr->op_ptr;
}

/* -------------------------------------------------------------------- */

/*
 * ET construction
 */
V_LOCAL et_node_p
etNewConstantLeaf(vec_t constant)
{
	et_node_p self = (et_node_p)memAllocZ(sizeof(et_node_t));
	self->var.constant = constant;
	self->op_ptr = op_const;
	return self;
}

V_LOCAL et_node_p
etNewVariableLeaf(int var_idx)
{
	et_node_p self = (et_node_p)memAllocZ(sizeof(et_node_t));
	self->var.idx = var_idx;
	self->op_ptr = op_var;
	return self;
}

V_LOCAL et_node_p
etNewNode(et_op_t op_ptr)
{
	et_node_p self = (et_node_p)memAllocZ(sizeof(et_node_t));
	self->var.idx = -1;
	self->op_ptr = op_ptr;
	return self;
}

V_LOCAL void
etDestroyTree(et_node_p root)
{
	if(root->left)
		etDestroyTree(root->left);
	if(root->right)
		etDestroyTree(root->right);
	if(root->control)
		etDestroyTree(root->control);
	memFree(root);
}

/*
 * etConstructTreeFromRPN()
 *
 * Using 'scanner' as a lexer, constructs expression tree
 * from its reverse polish notation. All known variables (but
 * _not_ built-in _constants_) had to be stored in 'var_ids'
 * _prior_ to this call.
 */

#define ET_ENABLE_OPTIMIZATION
/*
#define ET_TRACES
*/

V_LOCAL et_node_p
etConstructTreeFromRPN(vhandle_t scanner,u8_stack_p var_ids)
{
	CALLER_NAME(etConstructTreeFromRPN)
	int var_idx;
	vec_t constant;
	et_op_t op_ptr;
	char *token;
	et_node_p self, left, right, control;
	vp_stack_t nodes = TYPED_STACK_EMPTY_INITIALIZER;
	vboolean expect_token = true;
	vboolean parse_error = false;
	for(;;)
	{
		token = lexNext(scanner);
		if(!token || STR_EQ(token,")"))
		{
			if(expect_token)
			{
				SYN_EXPECTED(scanner,"<variable> | <constant> | <operator>");
				parse_error = true;
			}
			break;
		}
		if(STR_EQ(token,",")) /* delimiter */
		{
			if(expect_token)
			{
				SYN_EXPECTED(scanner,"<variable> | <constant> | <operator>");
				parse_error = true;
				break;
			}
			expect_token = true;
			continue;
		}
		expect_token = false;
		if(etFindIdentifier(token,var_ids,&var_idx)) /* variable */
		{
#ifdef  ET_TRACES
			conlPrintf(LOG_TRACE,"Variable: '%s'.\n",token);
#endif  /* ET_TRACES */
			vpsPush(&nodes,etNewVariableLeaf(var_idx));
			continue;
		}
		if(strToFloat(token,&constant)) /* direct constant */
		{
#ifdef  ET_TRACES
			conlPrintf(LOG_TRACE,"Direct constant: '%s'.\n",token);
#endif  /* ET_TRACES */
			vpsPush(&nodes,etNewConstantLeaf(constant));
			continue;
		}
		op_ptr = etFindUnaryOp(token);
		if(op_ptr)
		{
			/* unary operation */
			if(vpsCount(&nodes) < 1)
			{
				/* stack underflow */
				PARSE_ERROR(scanner,"stack underflow");
				parse_error = true;
				break;
			}
#ifdef  ET_TRACES
			conlPrintf(LOG_TRACE,"Unary operation: '%s'.\n",token);
#endif  /* ET_TRACES */
			/* get operand */
			vpsPop(&nodes,(void_p*)&left);
#ifdef  ET_ENABLE_OPTIMIZATION
			/* if it is a constant, pre-evaluate the result */
			if(left->op_ptr == op_const)
			{
				et_node_t temp;
				temp.left = left;
				temp.op_ptr = op_ptr;
				constant = ET_EVAL(&temp,NULL);
#ifdef  ET_TRACES
				conlPrintf(LOG_TRACE,"Unary optimization: '%s' of '%g' is '%g'.\n",
					token,ET_EVAL(left,NULL),constant);
#endif  /* ET_TRACES */
				self = etNewConstantLeaf(constant);
				etDestroyTree(left);
			}
			else
			{
#endif  /* ET_ENABLE_OPTIMIZATION */
				self = etNewNode(op_ptr);
				/* connect operand to _left_ node */
				self->left = left;
#ifdef  ET_ENABLE_OPTIMIZATION
			}
#endif  /* ET_ENABLE_OPTIMIZATION */
			vpsPush(&nodes,self);
			continue;
		}
		op_ptr = etFindBinaryOp(token);
		if(op_ptr)
		{
			/* binary operation */
			if(vpsCount(&nodes) < 2)
			{
				/* stack underflow */
				PARSE_ERROR(scanner,"stack underflow");
				parse_error = true;
				break;
			}
#ifdef  ET_TRACES
			conlPrintf(LOG_TRACE,"Binary operation: '%s'.\n",token);
#endif  /* ET_TRACES */
			/* get operands */
			vpsPop(&nodes,(void_p*)&right);
			vpsPop(&nodes,(void_p*)&left);
#ifdef  ET_ENABLE_OPTIMIZATION
			/* if both of them are constants, pre-evaluate the result */
			if(left->op_ptr == op_const && right->op_ptr == op_const)
			{
				et_node_t temp;
				temp.left = left;
				temp.right = right;
				temp.op_ptr = op_ptr;
				constant = ET_EVAL(&temp,NULL);
#ifdef  ET_TRACES
				conlPrintf(LOG_TRACE,
					"Binary optimization: '%s' of (%g,%g) is '%g'.\n",token,
					ET_EVAL(left,NULL),ET_EVAL(right,NULL),constant);
#endif  /* ET_TRACES */
				self = etNewConstantLeaf(constant);
				etDestroyTree(left);
				etDestroyTree(right);
			}
			else
			{
#endif  /* ET_ENABLE_OPTIMIZATION */
				self = etNewNode(op_ptr);
				/* connect operands */
				self->left = left;
				self->right = right;
#ifdef  ET_ENABLE_OPTIMIZATION
			}
#endif  /* ET_ENABLE_OPTIMIZATION */
			vpsPush(&nodes,self);
			continue;
		}
		op_ptr = etFindTernaryOp(token);
		if(op_ptr)
		{
			/* ternary operation */
			if(vpsCount(&nodes) < 3)
			{
				/* stack underflow */
				PARSE_ERROR(scanner,"stack underflow");
				parse_error = true;
				break;
			}
#ifdef  ET_TRACES
			conlPrintf(LOG_TRACE,"Ternary operation: '%s'.\n",token);
#endif  /* ET_TRACES */
			/* get operands */
			vpsPop(&nodes,(void_p*)&control);
			vpsPop(&nodes,(void_p*)&right);
			vpsPop(&nodes,(void_p*)&left);
#ifdef  ET_ENABLE_OPTIMIZATION
			/* if both of them are constants, pre-evaluate the result */
			if(left->op_ptr == op_const && 
				right->op_ptr == op_const &&
				control->op_ptr == op_const)
			{
				et_node_t temp;
				temp.left = left;
				temp.right = right;
				temp.control = control;
				temp.op_ptr = op_ptr;
				constant = ET_EVAL(&temp,NULL);
#ifdef  ET_TRACES
				conlPrintf(LOG_TRACE,
					"Ternary optimization: '%s' of (%g,%g,%g) is '%g'.\n",
					token,
					ET_EVAL(left,NULL),
					ET_EVAL(right,NULL),
					ET_EVAL(control,NULL),
					constant);
#endif  /* ET_TRACES */
				self = etNewConstantLeaf(constant);
				etDestroyTree(left);
				etDestroyTree(right);
				etDestroyTree(control);
			}
			else
			{
#endif  /* ET_ENABLE_OPTIMIZATION */
				self = etNewNode(op_ptr);
				/* connect operands */
				self->left = left;
				self->right = right;
				self->control = control;
#ifdef  ET_ENABLE_OPTIMIZATION
			}
#endif  /* ET_ENABLE_OPTIMIZATION */
			vpsPush(&nodes,self);
			continue;
		}
		if(etFindBuiltInConstant(token,&constant)) /* built-in constant */
		{
#ifdef  ET_TRACES
			conlPrintf(LOG_TRACE,"Built-in constant: '%s' = '%g'.\n",
				token,constant);
#endif  /* ET_TRACES */
			vpsPush(&nodes,etNewConstantLeaf(constant));
			continue;
		}
		SYN_EXPECTED(scanner,"<variable> | <constant> | <operator>");
		parse_error = true;
		break;
	}
	if(parse_error)
	{
		while(vpsPop(&nodes,(void_p)&self))
			etDestroyTree(self);
		return NULL;
	}
	if(vpsCount(&nodes) > 1)
	{
		conlPrintf(LOG_WARNING,"%s(): %u unused operand(s).\n",
			caller_name,vpsCount(&nodes) - 1);
	} else if(vpsCount(&nodes) == 0)
	{
		PARSE_ERROR(scanner,"empty expression");
		return NULL;
	}
	vpsPop(&nodes,(void_p)&self);
	while(vpsPop(&nodes,(void_p)&left))
		etDestroyTree(left);
	vpsClear(&nodes);
	return self;
}

V_LOCAL et_node_p
etParseExpression(vhandle_t scanner,u8_stack_p var_ids)
{
	CALLER_NAME(etParseExpression)
	vec_t val;
	et_node_p expr;
	if(lexNextIs(scanner,"("))
	{
		expr = etConstructTreeFromRPN(scanner,var_ids);
		if(expr && !lexLastWas(scanner,")"))
		{
			SYN_EXPECTED(scanner,")");
			etDestroyTree(expr);
			expr = NULL;
		}
	}
	else
	{
		if(strToFloat(lexLast(scanner),&val)) /* constant expression */
			expr = etNewConstantLeaf(val);
		else
		{
			SYN_EXPECTED(scanner,"(");
			return NULL;
		}
	}
	return expr;
}

/*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Part 2: statements.
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

typedef struct bcode_op_s
{
	/*
	 * Execution of the compaund statement continues while
	 * op[]->operate() routines return 'true'. Currently,
	 * only the 'return' operator returns 'false'.
	 */
	vboolean    (*operate)(struct bcode_op_s *self,vec_p variables);
	void        (*destroy)(struct bcode_op_s *self);
}   bcode_op_t, *bcode_op_p;

#define SYN_KEYWORD         0   /* any operator except assignment */
#define SYN_VARIABLE        1   /* assignment */

typedef struct op_start_id_s
{
	u32_t   token_type;
	char    *token_value;
}   op_start_id_t, *op_start_id_p;

typedef struct op_parse_entry_s
{
	vboolean        (*parse)(vhandle_t scanner,
		u8_stack_p var_ids, bcode_op_p *op_ret);
	op_start_id_t   op_syntax;
	char            *op_syntaxt_desc;
}   op_parse_entry_t, *op_parse_entry_p;

V_LOCAL vboolean
bcodeExecuteCompaund(vp_stack_p body,vec_p variables)
{
	bcode_op_p ptr;
	u32_t i, nops = vpsCount(body);
	for(i = 0;i < nops;i++)
	{
		ptr = (bcode_op_p)vpsGet(body,i);
		if(!ptr->operate(ptr,variables))
			return false;
	}
	return true;
}

V_LOCAL void
bcodeDestroyCompaund(vp_stack_p body)
{
	bcode_op_p ptr;
	while(vpsPop(body,(void_p)&ptr))
		ptr->destroy(ptr);
	vpsClear(body);
}

V_LOCAL vboolean
bcodeParseStatement(vhandle_t scanner,
	op_parse_entry_p known_ops,
	u8_stack_p var_ids,
	bcode_op_p *op_ret)
{
	CALLER_NAME(bcodeParseStatement)
	int idx;
	vboolean res = false;
	op_parse_entry_p ptr;
	/* assume first token of statement is already read in */
	for(ptr = known_ops;ptr->parse;ptr++)
	{
		switch(ptr->op_syntax.token_type)
		{
			case SYN_KEYWORD:
				if(lexLastWasEx(scanner,ptr->op_syntax.token_value,STR_COMP))
				{
					res = ptr->parse(scanner,var_ids,op_ret);
					goto outer_break;
				}
				continue;
			case SYN_VARIABLE:
				if(etFindIdentifier(lexLast(scanner),var_ids,&idx))
				{
					res = ptr->parse(scanner,var_ids,op_ret);
					goto outer_break;
				}
				continue;
			default:
				V_ASSERT(0); /* shouldn't happen */
				return false;
		}
	}
	PARSE_ERROR(scanner,"unknown statement");
	return false;
outer_break:;
	if(res)
		return true;
	if(*op_ret)
		(*op_ret)->destroy(*op_ret);
	*op_ret = NULL;
	return false;
}

V_LOCAL vboolean
bcodeParseStatements(vhandle_t scanner,
	op_parse_entry_p known_ops,
	u8_stack_p var_ids,
	vp_stack_p body)
{
	bcode_op_p op;
	vpsClear(body);
	for(;;)
	{
		if(!lexNext(scanner))
			break;
		if(lexLastWas(scanner,"}"))
			return true;
		op = NULL;
		if(bcodeParseStatement(scanner,known_ops,var_ids,&op))
		{
			if(op)
				vpsPush(body,op);
			continue;
		}
		break;
	}
	bcodeDestroyCompaund(body);
	return false;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/*
 * There is two kind of operators: declaration of variable ans assignment.
 * Former starts by keyword 'var' and requires NO operations (all that have
 * to be done is to expand 'var_ids').
 */
V_LOCAL vboolean
var_parse(vhandle_t scanner,u8_stack_p var_ids,bcode_op_p *op_ret)
{
	u32_t limit;
	CALLER_NAME(var_parse)
	*op_ret = NULL; /* we must handle this special case later */
	limit = u8sCount(var_ids);
	/* assume keyword 'var' already read in */
	for(;;)
	{
		if(!lexNext(scanner))
		{
			SYN_EXPECTED(scanner,"<identifier>");
			u8sResize(var_ids,limit);
			return false;
		}
		if(strlen(lexLast(scanner)) >= MAX_ID)
		{
			PARSE_ERROR(scanner,"identifier too long");
			u8sResize(var_ids,limit);
			return false;
		}
		etNewIdentifier(var_ids,lexLast(scanner));
		if(lexNextIs(scanner,","))
			continue;
		if(lexLastWas(scanner,";"))
			return true;
		break;
	}
	SYN_EXPECTED(scanner,",|;");
	u8sResize(var_ids,limit);
	return false;
}

#define VAR_PARSE_ENTRY \
	{var_parse,{SYN_KEYWORD,"var"}, \
	"var <identifier>[,<identifier>,...,<identifier>];"}

/*
 * Assignment operator
 */
typedef struct bcode_op_assign_s
{
	bcode_op_t  head;
	int         left_side; /* index of left side variable */
	et_node_p   right_side; /* tree of right-side expression */
}   bcode_op_assign_t, *bcode_op_assign_p;

V_LOCAL vboolean
assign_operate(struct bcode_op_s *self,vec_p variables)
{
	bcode_op_assign_p op = (bcode_op_assign_p)self;
	variables[op->left_side] = ET_EVAL(op->right_side,variables);
	variables[EXIT_CODE_OFFSET] = variables[op->left_side];
	return true;
}

V_LOCAL void
assign_destroy(struct bcode_op_s *self)
{
	bcode_op_assign_p op = (bcode_op_assign_p)self;
	etDestroyTree(op->right_side);
	memFree(op);
}

V_LOCAL vboolean
assign_parse(vhandle_t scanner,u8_stack_p var_ids,bcode_op_p *op_ret)
{
	CALLER_NAME(assign_parse)
	int ls_idx;
	et_node_p rs;
	bcode_op_assign_p op;
	*op_ret = NULL;
	/* assume the variable name already read in */
	if(!etFindIdentifier(lexLast(scanner),var_ids,&ls_idx))
	{
		SYN_UNDECLARED(scanner,lexLast(scanner));
		return false;
	}
	if(!lexNextIs(scanner,"="))
	{
		SYN_EXPECTED(scanner,"=");
		return false;
	}
	rs = etParseExpression(scanner,var_ids);
	if(!rs)
	{
		SYN_EXPECTED_EXPRESSION(scanner);
		return false;
	}
	if(!lexNextIs(scanner,";"))
	{
		SYN_EXPECTED(scanner,";");
		etDestroyTree(rs);
		return false;
	}
	op = (bcode_op_assign_p)memAllocZ(sizeof(bcode_op_assign_t));
	op->head.operate = assign_operate;
	op->head.destroy = assign_destroy;
	op->left_side = ls_idx;
	op->right_side = rs;
	*op_ret = (bcode_op_p)op;
	return true;
}

#define ASSIGN_PARSE_ENTRY \
	{assign_parse,{SYN_VARIABLE,NULL},"<variable> = <expression>;"}

/*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

V_LOCAL op_parse_entry_t known_statements[] = 
{
	VAR_PARSE_ENTRY,
	ASSIGN_PARSE_ENTRY,
	{NULL,{0,NULL},NULL}
};

/*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Part 3: functions.
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

typedef struct bcode_func_s
{
	u8_stack_t  var_ids;    /* identifiers, each of MAX_ID length */
	vec_p       variables;  /* memory for variables */
	vp_stack_t  body;       /* each of bcode_op_p */
}   bcode_func_t, *bcode_func_p;

V_LOCAL bcode_func_p
bcodeParseScript(vhandle_t scanner)
{
	bcode_func_p fn = (bcode_func_p)memAllocZ(sizeof(bcode_func_t));
	if(!etInitIdentifiers(&fn->var_ids))
	{
		memFree(fn);
		return NULL;
	}
	/* assume '{' already read in */
	if(!bcodeParseStatements(scanner,known_statements,&fn->var_ids,&fn->body))
	{
		memFree(fn);
		return NULL;
	}
	fn->variables = (vec_t*)memAllocZ(
		sizeof(vec_t) * (u8sCount(&fn->var_ids) / MAX_ID));
	return fn;
}

V_LOCAL vec_t
bcodeExecuteScript(bcode_func_p fn)
{
	bcodeExecuteCompaund(&fn->body,fn->variables);
	return fn->variables[EXIT_CODE_OFFSET];
}

V_LOCAL void
bcodeDestroyScript(bcode_func_p fn)
{
	bcodeDestroyCompaund(&fn->body);
	u8sClear(&fn->var_ids);
	memFree(fn->variables);
	memFree(fn);
}

/*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

int main()
{
	int row_idx, col_idx;
	bcode_func_p fn;
	vhandle_t scanner;
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	scanner = lexParseFile("test.musl","(),;");
	assert(scanner);
	fn = bcodeParseScript(scanner);
	lexDestroy(scanner);
	assert(fn);
	bcodeExecuteScript(fn);
	
	etFindBuiltInVariable("$ROW",&row_idx);
	etFindBuiltInVariable("$COLUMN",&col_idx);
	/*
	{
		int r, c;
		
		for(r = 0; r < 
		etc.
	}
	*/
	fn->variables[row_idx] = 10;
	fn->variables[col_idx] = 0;
	
	printf("Result: '%g'\n",fn->variables[EXIT_CODE_OFFSET]);
	bcodeDestroyScript(fn);
	memPrintStats();
	return 0;
}
