#include <v_lib.h>

/*
 * This is a realization of USL ver. 0.01.
 * USL stands for Ugly Scripting Language.
 */

#define STR_COMP        strcasecmp
#define STR_EQ(s1,s2)   (!STR_COMP(s1,s2))
#define MAX_ID          64

#ifdef  CHECKED_BUILD

#define PARSE_ERROR(scanner,desc) \
	conlPrintf(LOG_ERROR, \
		"%s(): error in \"%s\", line %u: %s (got \"%s\").\n", \
		caller_name,lexSourceName(scanner), \
		lexLastLine(scanner),desc,strSafe(lexLast(scanner)))

#define SYN_EXPECTED(scanner,desc) \
	conlPrintf(LOG_ERROR, \
		"%s(): error in \"%s\", line %u: expected \"%s\", got \"%s\".\n", \
		caller_name,lexSourceName(scanner), \
		lexLastLine(scanner),desc,strSafe(lexLast(scanner)))

#define SYN_UNDECLARED(scanner,desc) \
	conlPrintf(LOG_ERROR, \
		"%s(): error in \"%s\", line %u: \"%s\" undeclared.\n", \
		caller_name,lexSourceName(scanner),lexLastLine(scanner),desc)

#else   /* !CHECKED_BUILD */

#define PARSE_ERROR(scanner,desc)
#define SYN_EXPECTED(scanner,desc)
#define SYN_UNDECLARED(scanner,id_name)

#endif  /* !CHECKED_BUILD */

#define SYN_EXPECTED_EXPRESSION(scanner)    SYN_EXPECTED(scanner,"expression")

/*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Part 1: expressons.
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

/*
 * The first position in variables' array reserved for exit code.
 */
#define EXIT_CODE_OFFSET    0
#define EXIT_CODE_NAME      "%" /* mimic Maple syntax */

typedef struct et_built_in_var_s
{
	char    *var_name;
	int     var_offset;
	char    *var_desc;
}   et_built_in_var_t, *et_built_in_var_p;

typedef struct et_built_in_const_s
{
	char    *const_name;
	vec_t   const_value;
	char    *const_desc;
}   et_built_in_const_t, *et_built_in_const_p;

V_LOCAL et_built_in_var_t known_built_in_vars[] =
{
	{EXIT_CODE_NAME,EXIT_CODE_OFFSET,"exit code of last operator"},
/*
	{"$CELL",EXIT_CODE_OFFSET+1,"value of current cell"},
	{"$ROW",EXIT_CODE_OFFSET+2,"current row"},
	{"$COLUMN",EXIT_CODE_OFFSET+3,"current column"},
*/
	{NULL,-1,NULL}
};

V_LOCAL et_built_in_const_t known_built_in_consts[] =
{
	{"Pi",M_PI,"built-in: Pi"},
/*
	{"$MAX_ROW",31,"maximal index of row"},
	{"$MAX_COLUMN",63,"maximal index of column"},
*/
	{NULL,0,NULL}
};

int
etFirstBuiltInVariableOffset(void)
{
	et_built_in_var_p ptr = known_built_in_vars;
	if(ptr->var_name)
		return ptr->var_offset;
	return -1;
}

int
etLastBuiltInVariableOffset(void)
{
	et_built_in_var_p ptr = known_built_in_vars;
	if(!ptr->var_name)
		return -1;
	while(ptr->var_name)
		ptr++;
	ptr--;
	return ptr->var_offset;
}

vboolean
etFindBuiltInVariable(char *var_name,int *idx_ret)
{
	et_built_in_var_p ptr = known_built_in_vars;
	while(ptr->var_name)
	{
		if(STR_EQ(var_name,ptr->var_name))
		{
			if(idx_ret)
				*idx_ret = ptr->var_offset;
			return true;
		}
		ptr++;
	}
	return false;
}

vboolean
etFindBuiltInConstant(char *const_name,vec_t *const_ret)
{
	et_built_in_const_p ptr = known_built_in_consts;
	while(ptr->const_name)
	{
		if(STR_EQ(const_name,ptr->const_name))
		{
			if(const_ret)
				*const_ret = ptr->const_value;
			return true;
		}
		ptr++;
	}
	return false;
}

/* -------------------------------------------------------------------- */

V_LOCAL void
etNewIdentifier(u8_stack_p var_ids,char *var_name)
{

	char id[MAX_ID];
	memZero(id,MAX_ID);
	strCopy(id,var_name,MAX_ID);
	u8sPushBuffer(var_ids,(u8_p)id,MAX_ID);
	return;
}

V_LOCAL vboolean
etInitIdentifiers(u8_stack_p var_ids)
{
	int idx = 0;
	et_built_in_var_p ptr = known_built_in_vars;
	u8sInit(var_ids);
	/* add built-in variables */
	while(ptr->var_name)
	{
		/* _must_ be ordered */
		if(idx != ptr->var_offset)
		{
#ifdef  CHECKED_BUILD
			conlPrintf(LOG_ERROR,
				"etInitIdentifiers(): unordered built-in variable '%s'.\n",
				ptr->var_name);
#endif  /* CHECKED_BUILD */
			u8sClear(var_ids);
			return false;
		}
		etNewIdentifier(var_ids,ptr->var_name);
		ptr++;
		idx++;
	}
	return true;
}

V_LOCAL vboolean
etFindIdentifier(char *var_name,u8_stack_p var_ids,int *idx_ret)
{
	char *ptr;
	u32_t i, len;
	len = u8sCount(var_ids) / MAX_ID;
	ptr = (char*)u8sGetStorage(var_ids);
	for(i = 0;i < len;i++,ptr += MAX_ID)
	{
		if(STR_EQ(var_name,ptr))
		{
			if(idx_ret)
				*idx_ret = i;
			return true;
		}
	}
	return false;
}

struct et_node_s;

typedef vec_t (*et_op_t)(struct et_node_s *self,vec_t *vars);

typedef struct et_node_s
{
	union
	{
		int     idx;
		vec_t   constant;
	} var;
	et_op_t     op_ptr;
	struct      et_node_s *left, *right;
}   et_node_t, *et_node_p;

#define ET_EVAL(root,vars)  ((root)->op_ptr(root,vars))

/*
 * Constants and variables
 */
#include <../source/PAL/argsused.h>
V_LOCAL vec_t
op_const(et_node_p root,vec_t *vars)
{
	return root->var.constant;
}

V_LOCAL vec_t
op_var(et_node_p root,vec_t *vars)
{
	return vars[root->var.idx];
}

/*
 * Unary operations
 */
V_LOCAL vec_t
op_neg(et_node_p root,vec_t *vars)
{
	return -ET_EVAL(root->left,vars);
}

V_LOCAL vec_t
op_sin(et_node_p root,vec_t *vars)
{
	return (vec_t)sin(ET_EVAL(root->left,vars));
}

V_LOCAL vec_t
op_cos(et_node_p root,vec_t *vars)
{
	return (vec_t)cos(ET_EVAL(root->left,vars));
}

V_LOCAL vec_t
op_exp(et_node_p root,vec_t *vars)
{
	return (vec_t)exp(ET_EVAL(root->left,vars));
}

V_LOCAL vec_t
op_ln(et_node_p root,vec_t *vars)
{
	return (vec_t)log(ET_EVAL(root->left,vars));
}

/*
 * Binary operations
 */
V_LOCAL vec_t
op_add(et_node_p root,vec_t *vars)
{
	return ET_EVAL(root->left,vars) + ET_EVAL(root->right,vars);
}

V_LOCAL vec_t
op_sub(et_node_p root,vec_t *vars)
{
	return ET_EVAL(root->left,vars) - ET_EVAL(root->right,vars);
}

V_LOCAL vec_t
op_mul(et_node_p root,vec_t *vars)
{
	return ET_EVAL(root->left,vars) * ET_EVAL(root->right,vars);
}

V_LOCAL vec_t
op_div(et_node_p root,vec_t *vars)
{
	return ET_EVAL(root->left,vars) / ET_EVAL(root->right,vars);
}

V_LOCAL vec_t
op_mod(et_node_p root,vec_t *vars)
{
	return (vec_t)fmod(ET_EVAL(root->left,vars),ET_EVAL(root->right,vars));
}

V_LOCAL vec_t
op_pow(et_node_p root,vec_t *vars)
{
	return pow(ET_EVAL(root->left,vars),ET_EVAL(root->right,vars));
}

/*
 * Pseudo-boolean operations.
 * Returns:
 *  - 'false': _less_ than 0.5 (currently 0);
 *  - 'true': _greater_ than 0.5 (currently 1).
 */
V_LOCAL vec_t
op_less(et_node_p root,vec_t *vars)
{
	return (ET_EVAL(root->left,vars) < ET_EVAL(root->right,vars)) ?
		1 : 0;
}

V_LOCAL vec_t
op_greater(et_node_p root,vec_t *vars)
{
	return (ET_EVAL(root->left,vars) > ET_EVAL(root->right,vars)) ?
		1 : 0;
}

V_LOCAL vec_t
op_equal(et_node_p root,vec_t *vars)
{
	return IsZero(ET_EVAL(root->left,vars) - ET_EVAL(root->right,vars)) ?
		1 : 0;
}

V_LOCAL vec_t
op_not_equal(et_node_p root,vec_t *vars)
{
	return IsZero(ET_EVAL(root->left,vars) - ET_EVAL(root->right,vars)) ?
		0 : 1;
}

/* -------------------------------------------------------------------- */

typedef struct et_op_entry_s
{
	char        *op_name;
	et_op_t     op_ptr;
	char        *op_desc;
}   et_op_entry_t, *et_op_entry_p;

V_LOCAL et_op_entry_t known_unary_operations[] =
{
	{"neg",op_neg,"negates tos[0] (unary minus)"},
	{"sin",op_sin,"computes sine of tos[0]"},
	{"cos",op_cos,"computes cosine of tos[0]"},
	{"exp",op_exp,"computes exponent of tos[0]"},
	{"ln",op_ln,"computes natural logarithm of tos[0]"},
	{NULL,NULL,NULL}
};

V_LOCAL et_op_entry_t known_binary_operations[] =
{
	{"+",op_add,"adds tos[1] to tos[0]"},
	{"-",op_sub,"subtracts tos[0] form tos[1]"},
	{"*",op_mul,"multiplies tos[1] by tos[0]"},
	{"/",op_div,"divides tos[1] by tos[0]"},
	{"%",op_mod,"remainder of tos[1] / tos[0]"},
	{"^",op_pow,"raises tos[1] to the power of tos[0]"},
	{"<",op_less,"returns 1 if tos[1] < tos[0]"},
	{">",op_greater,"returns 1 if tos[1] > tos[0]"},
	{"==",op_equal,"returns 1 if |tos[1] - tos[0]| < FLOAT_ZERO_EPSILON"},
	{"!=",op_not_equal,"returns 1 if |tos[1] - tos[0]| > FLOAT_ZERO_EPSILON"},
	{NULL,NULL,NULL}
};

et_op_t
etFindUnaryOp(char *op_name)
{
	et_op_entry_p ptr = known_unary_operations;
	while(ptr->op_name)
	{
		if(STR_EQ(op_name,ptr->op_name))
			break;
		ptr++;
	}
	return ptr->op_ptr;
}

et_op_t
etFindBinaryOp(char *op_name)
{
	et_op_entry_p ptr = known_binary_operations;
	while(ptr->op_name)
	{
		if(STR_EQ(op_name,ptr->op_name))
			break;
		ptr++;
	}
	return ptr->op_ptr;
}

/* -------------------------------------------------------------------- */

/*
 * ET construction
 */
V_LOCAL et_node_p
etNewConstantLeaf(vec_t constant)
{
	et_node_p self = (et_node_p)memAllocZ(sizeof(et_node_t));
	self->var.constant = constant;
	self->op_ptr = op_const;
	return self;
}

V_LOCAL et_node_p
etNewVariableLeaf(int var_idx)
{
	et_node_p self = (et_node_p)memAllocZ(sizeof(et_node_t));
	self->var.idx = var_idx;
	self->op_ptr = op_var;
	return self;
}

V_LOCAL et_node_p
etNewNode(et_op_t op_ptr)
{
	et_node_p self = (et_node_p)memAllocZ(sizeof(et_node_t));
	self->var.idx = -1;
	self->op_ptr = op_ptr;
	return self;
}

V_LOCAL void
etDestroyTree(et_node_p root)
{
	if(root->left)
		etDestroyTree(root->left);
	if(root->right)
		etDestroyTree(root->right);
	memFree(root);
}

/*
 * etConstructTreeFromRPN()
 *
 * Using 'scanner' as a lexer, constructs expression tree
 * from its reverse polish notation. All known variables (but
 * _not_ built-in _constants_) had to be stored in 'var_ids'
 * _prior_ to this call.
 */

#define ET_ENABLE_OPTIMIZATION
/*
#define ET_TRACES
*/

V_LOCAL et_node_p
etConstructTreeFromRPN(vhandle_t scanner,u8_stack_p var_ids)
{
	CALLER_NAME(etConstructTreeFromRPN)
	int var_idx;
	vec_t constant;
	et_op_t op_ptr;
	char *token;
	et_node_p self, left, right;
	vp_stack_t nodes = TYPED_STACK_EMPTY_INITIALIZER;
	vboolean expect_token = true;
	vboolean parse_error = false;
	for(;;)
	{
		token = lexNext(scanner);
		if(!token || STR_EQ(token,")"))
		{
			if(expect_token)
			{
				SYN_EXPECTED(scanner,"<variable> | <constant> | <operator>");
				parse_error = true;
			}
			break;
		}
		if(STR_EQ(token,",")) /* delimiter */
		{
			if(expect_token)
			{
				SYN_EXPECTED(scanner,"<variable> | <constant> | <operator>");
				parse_error = true;
				break;
			}
			expect_token = true;
			continue;
		}
		expect_token = false;
		if(etFindIdentifier(token,var_ids,&var_idx)) /* variable */
		{
#ifdef  ET_TRACES
			conlPrintf(LOG_TRACE,"Variable: '%s'.\n",token);
#endif  /* ET_TRACES */
			vpsPush(&nodes,etNewVariableLeaf(var_idx));
			continue;
		}
		if(strToFloat(token,&constant)) /* direct constant */
		{
#ifdef  ET_TRACES
			conlPrintf(LOG_TRACE,"Direct constant: '%s'.\n",token);
#endif  /* ET_TRACES */
			vpsPush(&nodes,etNewConstantLeaf(constant));
			continue;
		}
		op_ptr = etFindUnaryOp(token);
		if(op_ptr)
		{
			/* unary operation */
			if(vpsCount(&nodes) < 1)
			{
				/* stack underflow */
				PARSE_ERROR(scanner,"stack underflow");
				parse_error = true;
				break;
			}
#ifdef  ET_TRACES
			conlPrintf(LOG_TRACE,"Unary operation: '%s'.\n",token);
#endif  /* ET_TRACES */
			/* get operand */
			vpsPop(&nodes,(void_p*)&left);
#ifdef  ET_ENABLE_OPTIMIZATION
			/* if it is a constant, pre-evaluate the result */
			if(left->op_ptr == op_const)
			{
				et_node_t temp;
				temp.left = left;
				temp.op_ptr = op_ptr;
				constant = ET_EVAL(&temp,NULL);
#ifdef  ET_TRACES
				conlPrintf(LOG_TRACE,"Unary optimization: '%s' of '%g' is '%g'.\n",
					token,ET_EVAL(left,NULL),constant);
#endif  /* ET_TRACES */
				self = etNewConstantLeaf(constant);
				etDestroyTree(left);
			}
			else
			{
#endif  /* ET_ENABLE_OPTIMIZATION */
				self = etNewNode(op_ptr);
				/* connect operand to _left_ node */
				self->left = left;
#ifdef  ET_ENABLE_OPTIMIZATION
			}
#endif  /* ET_ENABLE_OPTIMIZATION */
			vpsPush(&nodes,self);
			continue;
		}
		op_ptr = etFindBinaryOp(token);
		if(op_ptr)
		{
			/* binary operation */
			if(vpsCount(&nodes) < 2)
			{
				/* stack underflow */
				PARSE_ERROR(scanner,"stack underflow");
				parse_error = true;
				break;
			}
#ifdef  ET_TRACES
			conlPrintf(LOG_TRACE,"Binary operation: '%s'.\n",token);
#endif  /* ET_TRACES */
			/* get operands */
			vpsPop(&nodes,(void_p*)&right);
			vpsPop(&nodes,(void_p*)&left);
#ifdef  ET_ENABLE_OPTIMIZATION
			/* if both of them are constants, pre-evaluate the result */
			if(left->op_ptr == op_const && right->op_ptr == op_const)
			{
				et_node_t temp;
				temp.left = left;
				temp.right = right;
				temp.op_ptr = op_ptr;
				constant = ET_EVAL(&temp,NULL);
#ifdef  ET_TRACES
				conlPrintf(LOG_TRACE,
					"Binary optimization: '%s' of (%g,%g) is '%g'.\n",token,
					ET_EVAL(left,NULL),ET_EVAL(right,NULL),constant);
#endif  /* ET_TRACES */
				self = etNewConstantLeaf(constant);
				etDestroyTree(left);
				etDestroyTree(right);
			}
			else
			{
#endif  /* ET_ENABLE_OPTIMIZATION */
				self = etNewNode(op_ptr);
				/* connect operands */
				self->left = left;
				self->right = right;
#ifdef  ET_ENABLE_OPTIMIZATION
			}
#endif  /* ET_ENABLE_OPTIMIZATION */
			vpsPush(&nodes,self);
			continue;
		}
		if(etFindBuiltInConstant(token,&constant)) /* built-in constant */
		{
#ifdef  ET_TRACES
			conlPrintf(LOG_TRACE,"Built-in constant: '%s' = '%g'.\n",
				token,constant);
#endif  /* ET_TRACES */
			vpsPush(&nodes,etNewConstantLeaf(constant));
			continue;
		}
		SYN_EXPECTED(scanner,"<variable> | <constant> | <operator>");
		parse_error = true;
		break;
	}
	if(parse_error)
	{
		while(vpsPop(&nodes,(void_p)&self))
			etDestroyTree(self);
		return NULL;
	}
	if(vpsCount(&nodes) > 1)
	{
		conlPrintf(LOG_WARNING,"%s(): %u unused operand(s).\n",
			caller_name,vpsCount(&nodes) - 1);
	} else if(vpsCount(&nodes) == 0)
	{
		PARSE_ERROR(scanner,"empty expression");
		return NULL;
	}
	vpsPop(&nodes,(void_p)&self);
	while(vpsPop(&nodes,(void_p)&left))
		etDestroyTree(left);
	vpsClear(&nodes);
	return self;
}

V_LOCAL et_node_p
etParseExpression(vhandle_t scanner,u8_stack_p var_ids)
{
	CALLER_NAME(etParseExpression)
	vec_t val;
	et_node_p expr;
	if(lexNextIs(scanner,"("))
	{
		expr = etConstructTreeFromRPN(scanner,var_ids);
		if(expr && !lexLastWas(scanner,")"))
		{
			SYN_EXPECTED(scanner,")");
			etDestroyTree(expr);
			expr = NULL;
		}
	}
	else
	{
		if(strToFloat(lexLast(scanner),&val)) /* constant expression */
			expr = etNewConstantLeaf(val);
		else
		{
			SYN_EXPECTED(scanner,"(");
			return NULL;
		}
	}
	return expr;
}

/*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Part 2: statements.
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

typedef struct bcode_op_s
{
	/*
	 * Execution of the compaund statement continues while
	 * op[]->operate() routines return 'true'. Currently,
	 * only the 'return' operator returns 'false'.
	 */
	vboolean    (*operate)(struct bcode_op_s *self,vec_p variables);
	void        (*destroy)(struct bcode_op_s *self);
}   bcode_op_t, *bcode_op_p;

#define SYN_KEYWORD         0   /* any operator except assignment */
#define SYN_VARIABLE        1   /* assignment */

typedef struct op_start_id_s
{
	u32_t   token_type;
	char    *token_value;
}   op_start_id_t, *op_start_id_p;

typedef struct op_parse_entry_s
{
	vboolean        (*parse)(vhandle_t scanner,
		struct op_parse_entry_s *known_ops,
		u8_stack_p var_ids,
		bcode_op_p *op_ret);
	op_start_id_t   op_syntax;
	char            *op_syntaxt_desc;
}   op_parse_entry_t, *op_parse_entry_p;

V_LOCAL vboolean
bcodeExecuteCompaund(vp_stack_p body,vec_p variables)
{
	bcode_op_p ptr;
	u32_t i, nops = vpsCount(body);
	for(i = 0;i < nops;i++)
	{
		ptr = (bcode_op_p)vpsGet(body,i);
		if(!ptr->operate(ptr,variables))
			return false;
	}
	return true;
}

V_LOCAL void
bcodeDestroyCompaund(vp_stack_p body)
{
	bcode_op_p ptr;
	while(vpsPop(body,(void_p)&ptr))
		ptr->destroy(ptr);
	vpsClear(body);
}

V_LOCAL vboolean
bcodeParseStatement(vhandle_t scanner,
	op_parse_entry_p known_ops,
	u8_stack_p var_ids,
	bcode_op_p *op_ret)
{
	CALLER_NAME(bcodeParseStatement)
	int idx;
	vboolean res = false;
	op_parse_entry_p ptr;
	/* assume first token of statement is already read in */
	if(lexLastWas(scanner,";")) /* empty */
	{
		*op_ret = NULL;
		return true;
	}
	for(ptr = known_ops;ptr->parse;ptr++)
	{
		switch(ptr->op_syntax.token_type)
		{
			case SYN_KEYWORD:
				if(lexLastWasEx(scanner,ptr->op_syntax.token_value,STR_COMP))
				{
					res = ptr->parse(scanner,known_ops,var_ids,op_ret);
					goto outer_break;
				}
				continue;
			case SYN_VARIABLE:
				if(etFindIdentifier(lexLast(scanner),var_ids,&idx))
				{
					res = ptr->parse(scanner,known_ops,var_ids,op_ret);
					goto outer_break;
				}
				continue;
			default:
				V_ASSERT(0); /* shouldn't happen */
				return false;
		}
	}
	PARSE_ERROR(scanner,"unknown statement");
	return false;
outer_break:;
	if(res)
		return true;
	if(*op_ret)
		(*op_ret)->destroy(*op_ret);
	*op_ret = NULL;
	return false;
}

V_LOCAL vboolean
bcodeParseCompaund(vhandle_t scanner,
	op_parse_entry_p known_ops,
	u8_stack_p var_ids,
	vp_stack_p body)
{
	bcode_op_p op;
	vpsClear(body);
	if(!lexNext(scanner))
		return false;
	if(lexLastWas(scanner,"{")) /* compaund statement */
	{
		for(;;)
		{
			op = NULL;
			if(lexNextIs(scanner,"}"))
				break;
			if(bcodeParseStatement(scanner,known_ops,var_ids,&op))
			{
				if(op)
					vpsPush(body,op);
				continue;
			}
			bcodeDestroyCompaund(body);
			return false;
		}
	}
	else /* single statement */
	{
		op = NULL;
		if(!bcodeParseStatement(scanner,known_ops,var_ids,&op))
			return false;
		if(op)
			vpsPush(body,op);
	}
	return true;
}

/* ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ */

/*
 * There is only one special case of operator: declaration of variable.
 * It starts by keyword 'var' and requires NO operations (all that have
 * to be done is to expand 'var_ids').
 */
#include <../source/PAL/argsused.h>
V_LOCAL vboolean
var_parse(vhandle_t scanner,op_parse_entry_p known_ops,
	u8_stack_p var_ids,bcode_op_p *op_ret)
{
	u32_t limit;
	CALLER_NAME(var_parse)
	*op_ret = NULL; /* we must handle this special case later */
	limit = u8sCount(var_ids);
	/* assume keyword 'var' already read in */
	for(;;)
	{
		if(!lexNext(scanner))
		{
			SYN_EXPECTED(scanner,"<identifier>");
			u8sResize(var_ids,limit);
			return false;
		}
		if(strlen(lexLast(scanner)) >= MAX_ID)
		{
			PARSE_ERROR(scanner,"identifier too long");
			u8sResize(var_ids,limit);
			return false;
		}
		etNewIdentifier(var_ids,lexLast(scanner));
		if(lexNextIs(scanner,","))
			continue;
		if(!lexLastWas(scanner,";"))
		{
			SYN_EXPECTED(scanner,";");
			u8sResize(var_ids,limit);
			return false;
		}
		break;
	}
	return true; 
}

#define VAR_PARSE_ENTRY \
	{var_parse,{SYN_KEYWORD,"var"}, \
	"var <identifier>[,<identifier>,...,<identifier>];"}

/*
 * Return operator
 */
typedef struct bcode_op_return_s
{
	bcode_op_t  head;
	et_node_p   right_side; /* tree of right-side expression */
}   bcode_op_return_t, *bcode_op_return_p;

V_LOCAL vboolean
return_operate(struct bcode_op_s *self,vec_p variables)
{
	bcode_op_return_p op = (bcode_op_return_p)self;
	variables[EXIT_CODE_OFFSET] = ET_EVAL(op->right_side,variables);
	return false;
}

V_LOCAL void
return_destroy(struct bcode_op_s *self)
{
	bcode_op_return_p op = (bcode_op_return_p)self;
	etDestroyTree(op->right_side);
	memFree(op);
}

V_LOCAL vboolean
return_parse(vhandle_t scanner,op_parse_entry_p known_ops,
	u8_stack_p var_ids,bcode_op_p *op_ret)
{
	CALLER_NAME(return_parse)
	et_node_p rs;
	bcode_op_return_p op;
	*op_ret = NULL;
	/* assume that keyword 'return' is already read in */
	rs = etParseExpression(scanner,var_ids);
	if(!rs)
	{
		SYN_EXPECTED_EXPRESSION(scanner);
		return false;
	}
	if(!lexNextIs(scanner,";"))
	{
		SYN_EXPECTED(scanner,";");
		etDestroyTree(rs);
		return false;
	}
	op = (bcode_op_return_p)memAllocZ(sizeof(bcode_op_return_t));
	op->head.operate = return_operate;
	op->head.destroy = return_destroy;
	op->right_side = rs;
	*op_ret = (bcode_op_p)op;
	return true;
}

#define RETURN_PARSE_ENTRY \
	{return_parse,{SYN_KEYWORD,"return"},"return <expression>;"}

/*
 * Assignment operator
 */
typedef struct bcode_op_assign_s
{
	bcode_op_t  head;
	int         left_side; /* index of left side variable */
	et_node_p   right_side; /* tree of right-side expression */
}   bcode_op_assign_t, *bcode_op_assign_p;

V_LOCAL vboolean
assign_operate(struct bcode_op_s *self,vec_p variables)
{
	bcode_op_assign_p op = (bcode_op_assign_p)self;
	variables[op->left_side] = ET_EVAL(op->right_side,variables);
	variables[EXIT_CODE_OFFSET] = variables[op->left_side];
	return true;
}

V_LOCAL void
assign_destroy(struct bcode_op_s *self)
{
	bcode_op_assign_p op = (bcode_op_assign_p)self;
	etDestroyTree(op->right_side);
	memFree(op);
}

#include <../source/PAL/argsused.h>
V_LOCAL vboolean
assign_parse(vhandle_t scanner,op_parse_entry_p known_ops,
	u8_stack_p var_ids,bcode_op_p *op_ret)
{
	CALLER_NAME(assign_parse)
	int ls_idx;
	et_node_p rs;
	bcode_op_assign_p op;
	*op_ret = NULL;
	/* assume the variable name already read in */
	if(!etFindIdentifier(lexLast(scanner),var_ids,&ls_idx))
	{
		SYN_UNDECLARED(scanner,lexLast(scanner));
		return false;
	}
	if(!lexNextIs(scanner,"="))
	{
		SYN_EXPECTED(scanner,"=");
		return false;
	}
	rs = etParseExpression(scanner,var_ids);
	if(!rs)
	{
		SYN_EXPECTED_EXPRESSION(scanner);
		return false;
	}
	if(!lexNextIs(scanner,";"))
	{
		SYN_EXPECTED(scanner,";");
		etDestroyTree(rs);
		return false;
	}
	op = (bcode_op_assign_p)memAllocZ(sizeof(bcode_op_assign_t));
	op->head.operate = assign_operate;
	op->head.destroy = assign_destroy;
	op->left_side = ls_idx;
	op->right_side = rs;
	*op_ret = (bcode_op_p)op;
	return true;
}

#define ASSIGN_PARSE_ENTRY \
	{assign_parse,{SYN_VARIABLE,NULL},"<variable> = <expression>;"}

/*
 * Conditional operator
 */
typedef struct bcode_op_if_s
{
	bcode_op_t  head;
	et_node_p   control_expression; /* tree of control expression */
	vp_stack_t  true_clause;        /* each of bcode_op_p */
	vp_stack_t  false_clause;       /* each of bcode_op_p */
}   bcode_op_if_t, *bcode_op_if_p;

V_LOCAL vboolean
if_operate(struct bcode_op_s *self,vec_p variables)
{
	bcode_op_if_p op = (bcode_op_if_p)self;
	if(ET_EVAL(op->control_expression,variables) > 0.5)
		return bcodeExecuteCompaund(&op->true_clause,variables);
	return bcodeExecuteCompaund(&op->false_clause,variables);
}

V_LOCAL void
if_destroy(struct bcode_op_s *self)
{
	bcode_op_if_p op = (bcode_op_if_p)self;
	bcodeDestroyCompaund(&op->true_clause);
	bcodeDestroyCompaund(&op->false_clause);
	memFree(op);
}

V_LOCAL vboolean
if_parse(vhandle_t scanner,op_parse_entry_p known_ops,
	u8_stack_p var_ids,bcode_op_p *op_ret)
{
	CALLER_NAME(if_parse)
	vboolean has_else;
	bcode_op_if_p op = (bcode_op_if_p)memAllocZ(sizeof(bcode_op_if_t));
	op->head.operate = if_operate;
	op->head.destroy = if_destroy;
	*op_ret = NULL;
	/* assume the keyword 'if' is already read in */
	op->control_expression = etParseExpression(scanner,var_ids);
	if(!op->control_expression)
	{
		SYN_EXPECTED_EXPRESSION(scanner);
		memFree(op);
		return false;
	}
	if(!bcodeParseCompaund(scanner,known_ops,var_ids,&op->true_clause))
	{
		etDestroyTree(op->control_expression);
		memFree(op);
		SYN_EXPECTED(scanner,"<true_clause_statement>");
	}
	lexPush(scanner);
	has_else = lexNextIsEx(scanner,"else",STR_COMP);
	lexPop(scanner);
	if(has_else)
	{
		lexNext(scanner);
		if(!bcodeParseCompaund(scanner,known_ops,var_ids,&op->false_clause))
		{
			etDestroyTree(op->control_expression);
			bcodeDestroyCompaund(&op->true_clause);
			memFree(op);
			SYN_EXPECTED(scanner,"<false_clause_statement>");
			return false;
		}
	}
	else
		vpsClear(&op->false_clause);
	*op_ret = (bcode_op_p)op;
	return true;
}

#define IF_PARSE_ENTRY \
	{if_parse,{SYN_KEYWORD,"if"},"if <expression> <statement> [else <statement>]"}

/*
 * 'While' loop
 */
typedef struct bcode_op_while_s
{
	bcode_op_t  head;
	et_node_p   control_expression; /* tree of control expression */
	vp_stack_t  body;
}   bcode_op_while_t, *bcode_op_while_p;

V_LOCAL vboolean
while_operate(struct bcode_op_s *self,vec_p variables)
{
	bcode_op_while_p op = (bcode_op_while_p)self;
	while(ET_EVAL(op->control_expression,variables) > 0.5)
		if(!bcodeExecuteCompaund(&op->body,variables))
			return false;
	return true;
}

V_LOCAL void
while_destroy(struct bcode_op_s *self)
{
	bcode_op_while_p op = (bcode_op_while_p)self;
	etDestroyTree(op->control_expression);
	bcodeDestroyCompaund(&op->body);
	memFree(op);
}

V_LOCAL vboolean
while_parse(vhandle_t scanner,op_parse_entry_p known_ops,
	u8_stack_p var_ids,bcode_op_p *op_ret)
{
	CALLER_NAME(while_parse)
	bcode_op_while_p op = (bcode_op_while_p)memAllocZ(sizeof(bcode_op_while_t));
	op->head.operate = while_operate;
	op->head.destroy = while_destroy;
	*op_ret = NULL;
	/* assume the keyword 'while' is already read in */
	op->control_expression = etParseExpression(scanner,var_ids);
	if(!op->control_expression)
	{
		SYN_EXPECTED_EXPRESSION(scanner);
		memFree(op);
		return false;
	}
	if(!bcodeParseCompaund(scanner,known_ops,var_ids,&op->body))
	{
		etDestroyTree(op->control_expression);
		memFree(op);
		SYN_EXPECTED(scanner,"<while_body_statement>");
	}
	*op_ret = (bcode_op_p)op;
	return true;
}

#define WHILE_PARSE_ENTRY \
	{while_parse,{SYN_KEYWORD,"while"},"while <expression> <statement>"}

/*
 * 'For' loop
 */
typedef struct bcode_op_for_s
{
	bcode_op_t  head;
	int         control_variable;   /* index of control variable */
	vec_t       lower, upper, step; /* bounds of control variable */
	vp_stack_t  body;
}   bcode_op_for_t, *bcode_op_for_p;

V_LOCAL vboolean
for_operate(struct bcode_op_s *self,vec_p variables)
{
	bcode_op_for_p op = (bcode_op_for_p)self;
	vboolean direction = op->lower <= op->upper ? true : false;
	vec_p ctl = &variables[op->control_variable];
	if(direction)
	{
		for(*ctl = op->lower;*ctl <= op->upper;*ctl += op->step)
			if(!bcodeExecuteCompaund(&op->body,variables))
				return false;
	}
	else
	{
		for(*ctl = op->lower;*ctl >= op->upper;*ctl += op->step)
			if(!bcodeExecuteCompaund(&op->body,variables))
				return false;
	}
	return true;
}

V_LOCAL void
for_destroy(struct bcode_op_s *self)
{
	bcode_op_for_p op = (bcode_op_for_p)self;
	bcodeDestroyCompaund(&op->body);
	memFree(op);
}

V_LOCAL vboolean
for_parse(vhandle_t scanner,op_parse_entry_p known_ops,
	u8_stack_p var_ids,bcode_op_p *op_ret)
{
	CALLER_NAME(for_parse)
	int idx;
	vec_t lower, upper, step;
	bcode_op_for_p op;
	*op_ret = NULL;
	/* assume the keyword 'for' is already read in */
	if(!lexNext(scanner) ||
		!etFindIdentifier(lexLast(scanner),var_ids,&idx))
	{
		SYN_EXPECTED(scanner,"<identifier>");
		return false;
	}
	if(!lexNextIs(scanner,"="))
	{
		SYN_EXPECTED(scanner,"=");
		return false;
	}
	if(!strToFloat(lexNext(scanner),&lower))
	{
		SYN_EXPECTED(scanner,"<lower_limit_constant>");
		return false;
	}
	if(!lexNextIsEx(scanner,"to",STR_COMP))
	{
		SYN_EXPECTED(scanner,"to");
		return false;
	}
	if(!strToFloat(lexNext(scanner),&upper))
	{
		SYN_EXPECTED(scanner,"<upper_limit_constant>");
		return false;
	}
	if(!lexNextIsEx(scanner,"step",STR_COMP))
	{
		SYN_EXPECTED(scanner,"step");
		return false;
	}
	if(!strToFloat(lexNext(scanner),&step))
	{
		SYN_EXPECTED(scanner,"<step_constant>");
		return false;
	}
	op = (bcode_op_for_p)memAllocZ(sizeof(bcode_op_for_t));
	op->head.operate = for_operate;
	op->head.destroy = for_destroy;
	op->control_variable = idx;
	op->lower = lower;
	op->upper = upper;
	op->step = step;
	if(!bcodeParseCompaund(scanner,known_ops,var_ids,&op->body))
	{
		memFree(op);
		SYN_EXPECTED(scanner,"<for_body_statement>");
	}
	*op_ret = (bcode_op_p)op;
	return true;
}

#define FOR_PARSE_ENTRY \
	{for_parse,{SYN_KEYWORD,"for"}, \
		"for <identifier> = <start_const> to <end_const> " \
		"step <step_const> <statement>"}

/*
 * The fancy 'print' statement.
 * syntax:
 * "print [string|var],[string|var],...,[string|var];"
 */

typedef struct print_data_s
{
	vboolean	variable;
	union
	{
		int		var_idx;
		char	*str;
	}	data;
}	print_data_t, *print_data_p;

typedef struct bcode_op_print_s
{
	bcode_op_t  head;
	bstack_t	print_data;
}   bcode_op_print_t, *bcode_op_print_p;

V_LOCAL vboolean
print_operate(struct bcode_op_s *self,vec_p variables)
{
	u32_t i, nd;
	char buf[256];
	u8_stack_t str;
	print_data_p ptr;
	bcode_op_print_p op = (bcode_op_print_p)self;
	nd = stackCount(&op->print_data);
	ptr = (print_data_p)stackGetStorage(&op->print_data);
	u8sInit(&str);
	for(i = 0;i < nd;i++,ptr++)
	{
		if(ptr->variable)
		{
			sprintf(buf,"%g",variables[ptr->data.var_idx]);
			u8sPushBuffer(&str,(u8_p)buf,strlen(buf));
		}
		else
			u8sPushBuffer(&str,(u8_p)ptr->data.str,strlen(ptr->data.str));
	}
	u8sPush(&str,0);
	conPrintf("%s\n",u8sGetStorage(&str));
	u8sClear(&str);
	return true;
}

V_LOCAL void
print_destroy(struct bcode_op_s *self)
{
	u32_t i, nd;
	print_data_p ptr;
	bcode_op_print_p op = (bcode_op_print_p)self;
	nd = stackCount(&op->print_data);
	ptr = (print_data_p)stackGetStorage(&op->print_data);
	for(i = 0;i < nd;i++,ptr++)
		if(!ptr->variable)
			memFree(ptr->data.str);
	stackClear(&op->print_data);
	memFree(op);
}

V_LOCAL vboolean
print_parse(vhandle_t scanner,op_parse_entry_p known_ops,
	u8_stack_p var_ids,bcode_op_p *op_ret)
{
	CALLER_NAME(print_parse)
	print_data_t pd;
	bcode_op_print_p op;
	vboolean expect_something;
	*op_ret = NULL;
	op = (bcode_op_print_p)memAllocZ(sizeof(bcode_op_print_t));
	op->head.operate = print_operate;
	op->head.destroy = print_destroy;
	stackInit(&op->print_data,sizeof(print_data_t));
	expect_something = false;
	while(lexNext(scanner))
	{
		if(lexLastWas(scanner,";"))
		{
			if(expect_something)
			{
				SYN_EXPECTED(scanner,"<string> | <variable>");
				op->head.destroy((bcode_op_p)op);
				return false;
			}
			*op_ret = (bcode_op_p)op;
			return true;
		}
		if(lexLastWas(scanner,","))
		{
			expect_something = true;
			continue;
		}
		expect_something = false;
		if(etFindIdentifier(lexLast(scanner),var_ids,&pd.data.var_idx))
		{
			pd.variable = true;
			stackPush(&op->print_data,&pd);
			continue;
		}
		pd.variable = false;
		pd.data.str = strDup(lexLast(scanner));
		stackPush(&op->print_data,&pd);
	}
	PARSE_ERROR(scanner,"unexpected EOF");
	op->head.destroy((bcode_op_p)op);
	return false;
}

#define PRINT_PARSE_ENTRY \
	{print_parse,{SYN_KEYWORD,"print"}, \
		"print [string|var],[string|var],...,[string|var]"}

/*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

V_LOCAL op_parse_entry_t known_statements[] = 
{
	VAR_PARSE_ENTRY,
	RETURN_PARSE_ENTRY,
	ASSIGN_PARSE_ENTRY,
	IF_PARSE_ENTRY,
	WHILE_PARSE_ENTRY,
	FOR_PARSE_ENTRY,
	PRINT_PARSE_ENTRY,
	{NULL,{0,NULL},NULL}
};

/*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Part 3: functions.
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

typedef struct bcode_func_s
{
	u8_stack_t  var_ids;    /* identifiers, each of MAX_ID length */
	vec_p       variables;  /* memory for variables */
	vp_stack_t  body;       /* each of bcode_op_p */
}   bcode_func_t, *bcode_func_p;

bcode_func_p
bcodeParseFunction(vhandle_t scanner)
{
	bcode_func_p fn = (bcode_func_p)memAllocZ(sizeof(bcode_func_t));
	if(!etInitIdentifiers(&fn->var_ids))
	{
		memFree(fn);
		return NULL;
	}
	if(!bcodeParseCompaund(scanner,known_statements,&fn->var_ids,&fn->body))
	{
		memFree(fn);
		return NULL;
	}
	fn->variables = (vec_t*)memAllocZ(
		sizeof(vec_t) * (u8sCount(&fn->var_ids) / MAX_ID));
	return fn;
}

vec_t
bcodeExecuteFunction(bcode_func_p fn)
{
	bcodeExecuteCompaund(&fn->body,fn->variables);
	return fn->variables[EXIT_CODE_OFFSET];
}

void
bcodeDestroyFunction(bcode_func_p fn)
{
	bcodeDestroyCompaund(&fn->body);
	u8sClear(&fn->var_ids);
	memFree(fn->variables);
	memFree(fn);
}

/*
 * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 */

vec_t cf(vec_t val)
{
	vec_t f, x, limit;
	limit = val + 0.5;
	f = 1;
	x = 2;
	while(x < limit)
	{
		f = f * x;
		x = x + 1;
	}
	return f;
}

#define	LIM 10000

int main()
{
	bcode_func_p fn;
	vhandle_t scanner;
	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);
	scanner = lexParseFile("test.usl","(),;");
	assert(scanner);
	fn = bcodeParseFunction(scanner);
	lexDestroy(scanner);
	assert(fn);
	{
		u32_t i, tm;
		tm = timeTickCount();
		for(i = 0;i < LIM;i++)
		{
			cf(100);
		}
		tm = timeTickCount() - tm;
		printf("Direct: %u ms\n",tm);
		tm = timeTickCount();
		for(i = 0;i < LIM;i++)
		{
			fn->variables[EXIT_CODE_OFFSET] = 100;
			bcodeExecuteFunction(fn);
		}
		tm = timeTickCount() - tm;
		printf("Inerpreted: %u ms\n",tm);
	}
	bcodeDestroyFunction(fn);
	memPrintStats();
	return 0;
}
